"""
The spherical harmonic (SH) audio classes store audio data in the SH domain.
Please refer to the
:doc:`/theory/spherical_harmonic_definition` page for more general information.

The spherical harmonic audio classes are build upon the
:py:mod:`pyfar audio classes<pyfar.classes.audio>` and we recommend to get
familiar with these classes before continuing.

In addition to all functionality provided by the pyfar audio classes, the
spherical harmonic audio classes allow to store parameters defining the
spherical harmonics, these are the ``basis_type``, ``normalization``,
``channel_convention`` and the ``condon_shortley`` phase convention.
The last dimension of the channel shape must always match a valid number of
spherical harmonics, i.e. :math:`(N+1)^2`, where  :math:`N` is the spherical
harmonic order for which the audio data is created. The spherical harmonic
order of the data contained in the signal can be accessed through the property
``n_max``.

A SH signal can be created either directly

>>> import spharpy
>>> data = [[0, 0],  # data of first SH channel
...         [1, 1],  # data of second SH channel
...         [2, 2],  # data of third SH channel
...         [3, 4]]  # data of fourth SH channel
>>> sh_signal = spharpy.SphericalHarmonicSignal(
...     data, 44100, basis_type='real', normalization='N3D',
...     channel_convention='ACN', condon_shortley=False)

or from an SH definition

>>> # create a SH definition with default parameters
>>> definition = spharpy.SphericalHarmonicDefinition()
>>> sh_signal = spharpy.SphericalHarmonicSignal.from_definition(
...     definition, data, 44100)

Both examples create a first order SH signal with four SH channels and two time
samples at a sampling rate of 44.1 kHz.
"""
from pyfar import Signal, TimeData, FrequencyData
from pyfar.classes.audio import _Audio
from spharpy.spherical import renormalize, change_channel_convention
from spharpy.classes.sh import _SphericalHarmonicBase
import numpy as np
from abc import ABC


def _atleast_3d_first_dimension(data):
    """Ensure that data has at least 3 dimensions.
    Adds a singleton dimensions at the front if necessary.
    """

    data = np.atleast_2d(data)
    return data[np.newaxis, ...] if data.ndim < 3 else data


def _assert_valid_number_of_sh_channels(shape):
    """Check if the channel shape matches an integer spherical harmonic order.

    Parameters
    ----------
    shape : tuple, int
        Shape of the data array.

    Raises
    ------
    ValueError
        Raised if the number of spherical harmonic channels does not match
        (n_max + 1)^2 for an integer n_max.
    """

    sh_channels = shape[-2]
    n_max = np.sqrt(sh_channels)-1
    if n_max - int(n_max) != 0:
        raise ValueError(
            "Invalid number of spherical harmonic channels: "
            f"{sh_channels}. It must match (n_max + 1)^2.")


def _convert_to_standard_definition(
        data,
        normalization,
        channel_convention):
    """Convert data to the standard spherical harmonic definition.

    Parameters
    ----------
    data : numpy.ndarray
        Spherical harmonic data.
    normalization : str
        Normalization convention, either ``'N3D'``, ``'NM'``,
        ``'maxN'``, ``'SN3D'`` or ``'SNM'``. (maxN is only supported up
        to 3rd order)
    channel_convention : str
        Channel ordering convention, either ``'ACN'`` or ``'FuMa'``.
        (FuMa is only supported up to 3rd order)

    Returns
    -------
    numpy.ndarray
        Spherical harmonic data following the standard definition (N3D, ACN).
    """

    data = renormalize(
        data, channel_convention, normalization,
        "N3D", axis=-2)

    data = change_channel_convention(
        data, channel_convention, "ACN", axis=-2)

    return data


def _convert_from_standard_definition(
        data,
        normalization,
        channel_convention):
    """Convert data from standard definition to the desired one.

    Parameters
    ----------
    data : numpy.ndarray
        Spherical harmonic data using the standard definition (N3D, ACN).
    normalization : str
        Normalization convention, either ``'N3D'``, ``'NM'``,
        ``'maxN'``, ``'SN3D'`` or ``'SNM'``. (maxN is only supported up
        to 3rd order)
    channel_convention : str
        Channel ordering convention, either ``'ACN'`` or ``'FuMa'``.
        (FuMa is only supported up to 3rd order)

    Returns
    -------
    numpy.ndarray
        Spherical harmonic data according to the desired definition.
    """

    data = renormalize(
        data, "ACN", "N3D", normalization, axis=-2)

    data = change_channel_convention(
        data, "ACN", channel_convention, axis=-2)

    return data


class _SphericalHarmonicAudio(_Audio, _SphericalHarmonicBase, ABC):
    """
    Base class for spherical harmonics audio objects.

    This class extends the pyfar Audio class with all methods and
    properties required for spherical harmonic data and are common to the
    three sub-classes :py:class:`SphericalHarmonicTimeData`,
    :py:class:`SphericalHarmonicFrequencyData`, and
    :py:class:`SphericalHarmonicSignal`.

    Objects of this class contain spherical harmonic coefficients which are
    directly convertible between channel conventions ACN and FUMA, as
    well as the normalizations N3D, SN3D, or MaxN. The definition of
    the spherical harmonic basis functions is based on the scipy convention
    which includes the Condon-Shortley phase.


    Parameters
    ----------
    basis_type : str
        Type of spherical harmonic basis, either ``'complex'`` or
        ``'real'``.
    normalization : str
        Normalization convention, either ``'N3D'``, ``'NM'``, ``'SN3D'``,
        ``'SNM'``, or ``'maxN'``. ``'maxN'`` is only supported up to 3rd order.
    channel_convention : str
        Channel ordering convention, either ``'ACN'`` or ``'FuMa'``.
        ``'FuMa'`` is only supported up to 3rd order.
    condon_shortley : bool
        Whether to include the Condon-Shortley phase term. If ``True``,
        Condon-Shortley is included, if ``False`` it is not
        included. ``'auto'`` corresponds to ``True`` for complex `basis_type`
        and ``False`` for real `basis_type`.
    domain : ``'time'``, ``'freq'``, optional
        Domain of data. The default is ``'time'``
    comment : str
        A comment related to `data`. The default is ``None``.

    """

    def __init__(self, basis_type, normalization, channel_convention,
                 condon_shortley):

        _SphericalHarmonicBase.__init__(
            self,
            basis_type,
            normalization,
            channel_convention,
            condon_shortley)

    @property
    def n_max(self):
        """Get or set the spherical harmonic order."""
        return int(np.sqrt(self.cshape[-1])-1)

    @_SphericalHarmonicBase.basis_type.setter
    def basis_type(self, value):
        """Get or set the spherical harmonic basis type."""

        # Make sure that the basis type can only be set during initialization.
        # Changing it afterwards requires implementing the conversion between
        # real and complex-valued coefficients, which is possible but not yet
        # implemented.
        if self._basis_type is not None:
            raise AttributeError("Changing the basis_type is not yet "
                                 "supported.")
        else:
            _SphericalHarmonicBase.basis_type.fset(self, value)


class SphericalHarmonicTimeData(_SphericalHarmonicAudio, TimeData):
    """
    Create spherical harmonic audio object with time domain spherical
    harmonic coefficients and times.

    Objects of this class contain time data which is not directly convertible
    to the frequency domain, i.e., non-equidistant temporal sampling.

    Parameters
    ----------
    data : array, double
        Raw data in the time domain. The data should have at least 2
        dimensions, with the last dimension representing the time domain
        samples, the second to last the spherical harmonic coefficients,
        and any leading dimensions representing optional channels. Accordingly,
        the data should follow the 'C' memory layout, e.g. data of
        ``shape = (1, 4, 1024)`` has 1 channel with 4 spherical harmonic
        coefficients with 1024 samples each. The data can be ``int``,
        ``float`` or ``complex``. Data of type ``int`` is converted to
        ``float``.
    times : array, double
        Times in seconds at which the data is sampled. The number of times
        must match the size of the last dimension of `data`, i.e.,
        ``data.shape[-1]``.
    basis_type : str
        Type of spherical harmonic basis, either ``'complex'`` or
        ``'real'``.
    normalization : str
        Normalization convention, either ``'N3D'``, ``'NM'``,
        ``'maxN'``, ``'SN3D'`` or ``'SNM'``. (maxN is only supported up
        to 3rd order)
    channel_convention : str
        Channel ordering convention, either ``'ACN'`` or ``'FuMa'``.
        (FuMa is only supported up to 3rd order)
    condon_shortley : bool or str
        Whether to include the Condon-Shortley phase term. If ``True``,
        Condon-Shortley is included, if ``False`` it is not
        included. ``'auto'`` corresponds to ``True`` for complex `basis_type`
        and ``False`` for real `basis_type`.
    comment : str
        A comment related to `data`. The default is ``""``.
    is_complex : bool, optional
        A flag which indicates if the time data are real or complex-valued.
        The default is ``False``.
    """

    def __init__(self, data, times, basis_type, normalization,
                 channel_convention, condon_shortley, comment="",
                 is_complex=False):

        if not is_complex and basis_type == 'complex':
            raise ValueError(
                "Complex spherical harmonic basis requires "
                "complex time data. Set is_complex=True.")

        data = _atleast_3d_first_dimension(data)
        _assert_valid_number_of_sh_channels(data.shape)

        _SphericalHarmonicAudio.__init__(
            self, basis_type, normalization, channel_convention,
            condon_shortley)

        TimeData.__init__(self, data=data, times=times, comment=comment,
                          is_complex=is_complex)

    @classmethod
    def from_definition(
            cls, sh_definition, data, times, comment="", is_complex=False):
        r"""
        Create a SphericalHarmonicTimeData class object from
        SphericalHarmonicDefinition object, data, and times.

        Parameters
        ----------
        sh_definition : SphericalHarmonicDefinition
            The spherical harmonic definition.
        data : array, double
            Raw data in the time domain. The data should have at least 2
            dimensions, with the last dimension representing the time domain
            samples, the second to last the spherical harmonic coefficients,
            and any leading dimensions representing optional channels.
            Accordingly, the data should follow the 'C' memory layout, e.g.
            data of ``shape = (1, 4, 1024)`` has 1 channel with 4 spherical
            harmonic coefficients with 1024 samples each. The data can be
            ``int``, ``float`` or ``complex``. Data of type ``int`` is
            converted to ``float``.
        times : array, double
            Times in seconds at which the data is sampled. The number of times
            must match the size of the last dimension of `data`, i.e.,
            ``data.shape[-1]``.
        comment : str
            A comment related to `data`. The default is ``None``.
        is_complex : bool, optional
            A flag which indicates if the time data are real or complex-valued.
            The default is ``False``.
        """
        return cls(data, times,
                   basis_type=sh_definition.basis_type,
                   normalization=sh_definition.normalization,
                   channel_convention=sh_definition.channel_convention,
                   condon_shortley=sh_definition.condon_shortley,
                   comment=comment, is_complex=is_complex)

    @property
    def time(self):
        """Return or set the time data."""
        return _convert_from_standard_definition(TimeData.time.fget(self),
                                                 self.normalization,
                                                 self.channel_convention)

    @time.setter
    def time(self, value):
        """Return or set the time data."""
        value = _atleast_3d_first_dimension(value)
        _assert_valid_number_of_sh_channels(value.shape)

        value = _convert_to_standard_definition(
            value, self.normalization, self.channel_convention)
        TimeData.time.fset(self, value)


class SphericalHarmonicFrequencyData(_SphericalHarmonicAudio, FrequencyData):
    """
    Create spherical harmonic audio object with frequency domain spherical
    harmonic coefficients and frequencies.

    Objects of this class contain frequency data which is not directly
    convertible to the time domain, i.e., non-equidistantly spaced bins or
    incomplete spectra.

    Parameters
    ----------
    data : array, double
        Raw data in the frequency domain. The data should have at least
        2 dimensions, with the last dimension representing the frequency domain
        bins, the second to last the spherical harmonic coefficients,
        and any leading dimensions representing optional channels. Accordingly,
        the data should follow the 'C' memory layout, e.g. data of
        ``shape = (1, 4, 1024)`` has 1 channel with 4 spherical harmonic
        coefficients with 1024 frequency bins each. The data can be ``int``,
        ``float`` or ``complex``. Data of type ``int`` is converted to
        ``float``.
    frequencies : array, double
        Frequencies of the data in Hz. The number of frequencies must match
        the size of the last dimension of `data`, i.e., ``data.shape[-1]``.
    basis_type : str
        Type of spherical harmonic basis, either ``'complex'`` or
        ``'real'``.
    normalization : str
        Normalization convention, either ``'N3D'``, ``'NM'``, ``'SN3D'``,
        ``'SNM'``, or ``'maxN'``. ``'maxN'`` is only supported up to 3rd order.
    channel_convention : str
        Channel ordering convention, either ``'ACN'`` or ``'FuMa'``.
        ``'FuMa'`` is only supported up to 3rd order.
    condon_shortley : bool or str
        Whether to include the Condon-Shortley phase term. If ``True``,
        Condon-Shortley is included, if ``False`` it is not
        included. ``'auto'`` corresponds to ``True`` for complex `basis_type`
        and ``False`` for real `basis_type`.
    comment : str
        A comment related to `data`. The default is ``""``.
    """

    def __init__(self, data, frequencies, basis_type, normalization,
                 channel_convention, condon_shortley, comment=""):

        data = _atleast_3d_first_dimension(data)
        _assert_valid_number_of_sh_channels(data.shape)

        _SphericalHarmonicAudio.__init__(
            self, basis_type, normalization, channel_convention,
            condon_shortley)

        FrequencyData.__init__(self, data=data, frequencies=frequencies,
                               comment=comment)

    @classmethod
    def from_definition(
            cls, sh_definition, data, frequencies, comment=""):
        r"""
        Create a SphericalHarmonicFrequencyData class object from
        SphericalHarmonicDefinition object, data, and frequencies
        rate.

        Parameters
        ----------
        sh_definition : SphericalHarmonicDefinition
            The spherical harmonic definition.
        data : ndarray, double
            Raw data in the frequency domain. The data should have at least
            2 dimensions, with the last dimension representing the frequency
            domain bins, the second to last the spherical harmonic
            coefficients, and any leading dimensions representing optional
            channels. Accordingly, the data should follow the 'C' memory
            layout, e.g. data of ``shape = (1, 4, 1024)`` has 1 channel with 4
            spherical harmonic coefficients with 1024 frequency bins each. The
            data can be ``int``, ``float`` or ``complex``. Data of type
            ``int`` is converted to ``float``.
        frequencies : array, double
            Frequencies of the data in Hz. The number of frequencies must match
            the size of the last dimension of `data`, i.e., ``data.shape[-1]``.
        comment : str
            A comment related to `data`. The default is ``None``.
        """
        return cls(data, frequencies,
                   basis_type=sh_definition.basis_type,
                   normalization=sh_definition.normalization,
                   channel_convention=sh_definition.channel_convention,
                   condon_shortley=sh_definition.condon_shortley,
                   comment=comment)

    @property
    def freq(self):
        """Return or set the data in the frequency domain."""
        return _convert_from_standard_definition(FrequencyData.freq.fget(self),
                                                 self.normalization,
                                                 self.channel_convention)

    @freq.setter
    def freq(self, value):
        """Return or set the data in the frequency domain."""
        value = _atleast_3d_first_dimension(value)
        _assert_valid_number_of_sh_channels(value.shape)

        value = _convert_to_standard_definition(
            value, self.normalization, self.channel_convention)

        FrequencyData.freq.fset(self, value)


class SphericalHarmonicSignal(_SphericalHarmonicAudio, Signal):
    """
    Create audio object with spherical harmonic coefficients in time or
    frequency domain.

    Objects of this class contain spherical harmonic coefficients which are
    directly convertible between time and frequency domain (equally spaced
    samples and frequency bins), the channel conventions ACN and FuMa, as
    well as the normalizations N3D, SN3D, or MaxN, see [#]_. The definition of
    the spherical harmonics basis functions is based on the scipy convention
    which includes the Condon-Shortley phase, [#]_, [#]_.


    Parameters
    ----------
    data : ndarray, double
        Raw data of the spherical harmonics signal in the time or
        frequency domain. The data should have at least 2 dimensions, with
        the last dimension representing the time domain
        samples/frequency domain bins, the second to last the spherical
        harmonic coefficients, and any leading dimensions representing
        optional channels. Accordingly, the data should follow the 'C'
        memory layout, e.g. data of ``shape = (1, 4, 1024)`` has 1 channel
        with 4 spherical harmonic coefficients with 1024 samples or frequency
        bins each. Time data is converted to ``float``. Frequency is
        converted to ``complex`` and must be provided as single
        sided spectra, i.e., for all frequencies between 0 Hz and
        half the sampling rate.
    sampling_rate : double
        Sampling rate in Hz
    basis_type : str
        Type of spherical harmonic basis, either ``'complex'`` or
        ``'real'``.
    normalization : str
        Normalization convention, either ``'N3D'``, ``'NM'``, ``'SN3D'``,
        ``'SNM'``, or ``'maxN'``. ``'maxN'`` is only supported up to 3rd order.
    channel_convention : str
        Channel ordering convention, either ``'ACN'`` or ``'FuMa'``.
        ``'FuMa'`` is only supported up to 3rd order.
    condon_shortley : bool or str
        Whether to include the Condon-Shortley phase term. If ``True``,
        Condon-Shortley is included, if ``False`` it is not
        included. ``'auto'`` corresponds to ``True`` for complex `basis_type`
        and ``False`` for real `basis_type`.
    n_samples : int, optional
        Number of time domain samples. Required if domain is ``'freq'``.
        The default is ``None``, which assumes an even number of samples
        if the data is provided in the frequency domain.
    domain : ``'time'``, ``'freq'``, optional
        Domain of data. The default is ``'time'``
    fft_norm : str, optional
        The normalization of the Discrete Fourier Transform (DFT). Can be
        ``'none'``, ``'unitary'``, ``'amplitude'``, ``'rms'``, ``'power'``,
        or ``'psd'``. See :py:func:`~pyfar.dsp.fft.normalization`
        for more information. The default is ``'none'``, which is typically
        used for energy signals, such as impulse responses.
    comment : str
        A comment related to `data`. The default is ``None``.
    is_complex : bool, optional
        Specifies if the underlying time domain data are complex
        or real-valued. If ``True`` and `domain` is ``'time'``, the
        input data will be cast to complex. The default is ``False``.

    References
    ----------
    .. [#] F. Zotter, M. Frank, "Ambisonics A Practical 3D Audio Theory
            for Recording, Studio Production, Sound Reinforcement, and
            Virtual Reality", (2019), Springer-Verlag
    .. [#] B. Rafely, "Fundamentals of Spherical Array Processing", (2015),
            Springer-Verlag
    .. [#] E.G. Williams, "Fourier Acoustics", (1999), Academic Press

    """

    def __init__(self,
                 data,
                 sampling_rate,
                 basis_type,
                 normalization,
                 channel_convention,
                 condon_shortley,
                 n_samples=None,
                 domain='time',
                 fft_norm='none',
                 comment="",
                 is_complex=False):

        data = _atleast_3d_first_dimension(data)
        _assert_valid_number_of_sh_channels(data.shape)

        _SphericalHarmonicAudio.__init__(
            self, basis_type, normalization, channel_convention,
            condon_shortley)

        Signal.__init__(self, data=data, sampling_rate=sampling_rate,
                        n_samples=n_samples, domain=domain, fft_norm=fft_norm,
                        comment=comment, is_complex=is_complex)

    @classmethod
    def from_definition(
            cls, sh_definition, data, sampling_rate, domain='time',
            fft_norm='none', comment="", is_complex=False):
        r"""
        Create a SphericalHarmonicSignal class object from
        SphericalHarmonicDefinition object, data, and sampling
        rate.

        Parameters
        ----------
        sh_definition : SphericalHarmonicDefinition
            The spherical harmonic definition.
        data : ndarray, double
            Raw data of the spherical harmonic signal in the time or
            frequency domain. The data should have at least 2 dimensions, with
            the last dimension representing the time domain
            samples/frequency domain bins, the second to last the spherical
            harmonic coefficients, and any leading dimensions representing
            optional channels. Accordingly, the data should follow the 'C'
            memory layout, e.g. data of ``shape = (1, 4, 1024)`` has 1 channel
            with 4 spherical harmonic coefficients with 1024 samples or
            frequency bins each. Time data is converted to ``float``.
            Frequency is converted to ``complex`` and must be provided as
            single sided spectra, i.e., for all frequencies between 0 Hz and
            half the sampling rate.
        sampling_rate : double
            Sampling rate in Hz
        domain : ``'time'``, ``'freq'``, optional
            Domain of data. The default is ``'time'``
        fft_norm : str, optional
            The normalization of the Discrete Fourier Transform (DFT). Can be
            ``'none'``, ``'unitary'``, ``'amplitude'``, ``'rms'``, ``'power'``,
            or ``'psd'``. See :py:func:`~pyfar.dsp.fft.normalization`
            for more information. The default is ``'none'``, which is typically
            used for energy signals, such as impulse responses.
        comment : str
            A comment related to `data`. The default is ``None``.
        is_complex : bool, optional
            Specifies if the underlying time domain data are complex
            or real-valued. If ``True`` and `domain` is ``'time'``, the
            input data will be cast to complex. The default is ``False``.
        """
        return cls(data, sampling_rate,
                   basis_type=sh_definition.basis_type,
                   normalization=sh_definition.normalization,
                   channel_convention=sh_definition.channel_convention,
                   condon_shortley=sh_definition.condon_shortley,
                   domain=domain, fft_norm=fft_norm,
                   comment=comment, is_complex=is_complex)

    @property
    def freq(self):
        """Return or set the data in the frequency domain."""
        return _convert_from_standard_definition(Signal.freq.fget(self),
                                                 self.normalization,
                                                 self.channel_convention)

    @freq.setter
    def freq(self, value):
        """Return or set the data in the frequency domain."""
        value = _atleast_3d_first_dimension(value)
        _assert_valid_number_of_sh_channels(value.shape)

        value = _convert_to_standard_definition(
            value, self.normalization, self.channel_convention)

        Signal.freq.fset(self, value)

    @property
    def freq_raw(self):
        """Return or set the frequency domain data without normalization."""
        return _convert_from_standard_definition(Signal.freq_raw.fget(self),
                                                 self.normalization,
                                                 self.channel_convention)

    @freq_raw.setter
    def freq_raw(self, value):
        """Return or set the frequency domain data without normalization."""
        value = _atleast_3d_first_dimension(value)
        _assert_valid_number_of_sh_channels(value.shape)

        value = _convert_to_standard_definition(
            value, self.normalization, self.channel_convention)

        Signal.freq_raw.fset(self, value)

    @property
    def time(self):
        """Return or set the time data."""
        return _convert_from_standard_definition(Signal.time.fget(self),
                                                 self.normalization,
                                                 self.channel_convention)

    @time.setter
    def time(self, value):
        """Return or set the time data."""
        value = _atleast_3d_first_dimension(value)
        _assert_valid_number_of_sh_channels(value.shape)

        value = _convert_to_standard_definition(
            value, self.normalization, self.channel_convention)
        Signal.time.fset(self, value)



"""
Helper functions for coordinate operations.
"""

import numpy as np
from scipy.spatial import cKDTree, SphericalVoronoi


def coordinates2latlon(coordinates):
    r"""Transforms from Cartesian coordinates to Geocentric coordinates.

    .. math::

        h = \sqrt{x^2 + y^2 + z^2},

        \theta = \pi/2 - \arccos(\frac{z}{r}),

        \phi = \arctan(\frac{y}{x})

        -\pi/2 < \theta < \pi/2,

        -\pi < \phi < \pi

    where :math:`h` is the height, :math:`\theta` is the latitude angle
    and :math:`\phi` is the longitude angle

    Parameters
    ----------
    coordinates : ndarray, number
        Coordinates Object which cartesian coordinates are taken to convert to
        Geocentric coordinates

    Returns
    -------
    height : ndarray, number
        The radius is rendered as height information
    latitude : ndarray, number
        Geocentric latitude angle
    longitude : ndarray, number
        Geocentric longitude angle

    """
    x = coordinates.x
    y = coordinates.y
    z = coordinates.z
    height = np.sqrt(x**2 + y**2 + z**2)
    latitude = np.pi/2 - np.arccos(z/height)
    longitude = np.arctan2(y, x)
    return height, latitude, longitude


def spherical_voronoi(sampling, round_decimals=13, center=0.0):
    """Calculate a Voronoi diagram on the sphere for the given samplings
    points.

    Parameters
    ----------
    sampling : SamplingSphere
        Sampling points on a sphere
    round_decimals : int
        Number of decimals to be rounded to. Default is ``13``.
    center : double
        Center point of the voronoi diagram. Default is ``0.0``.

    Returns
    -------
    voronoi : SphericalVoronoi
        Spherical voronoi diagram as implemented in scipy.

    """
    points = sampling.cartesian
    points = points if points.shape[-1] == 3 else points.T
    radius = np.unique(np.round(sampling.radius, decimals=round_decimals))
    if len(radius) > 1:
        raise ValueError("All sampling points need to be on the \
                same radius.")
    voronoi = SphericalVoronoi(points, radius[0], center)

    return voronoi


def calculate_sampling_weights(sampling, round_decimals=12):
    """Calculate the sampling weights for numeric integration.

    Parameters
    ----------
    sampling : SamplingSphere
        Sampling points on a sphere
    round_decimals : int, optional
        Decimal precision used for ``sampling.radius``. All radii must be
        identical to compute the sampling weights. This can be used to ignore
        numerical noise on the radii. The default is ``12``.

    Returns
    -------
    weigths : ndarray, float
        Sampling weights

    """
    sv = spherical_voronoi(sampling, round_decimals=round_decimals)
    sv.sort_vertices_of_regions()

    unique_verts, idx_uni = np.unique(
        np.round(sv.vertices, decimals=10),
        axis=0,
        return_index=True)

    searchtree = cKDTree(unique_verts)
    if hasattr(sampling, 'csize'):
        area = np.zeros(sampling.csize, float)
    else:
        area = np.zeros(sampling.n_points, float)

    for idx, region in enumerate(sv.regions):
        _, idx_nearest = searchtree.query(sv.vertices[np.array(region)])
        mask_unique = np.sort(np.unique(idx_nearest, return_index=True)[1])
        mask_new = idx_uni[idx_nearest[mask_unique]]

        area[idx] = _poly_area(sv.vertices[mask_new])

    area = area / np.sum(area) * 4 * np.pi

    return area


def _unit_normal(a, b, c):
    x = np.linalg.det(
        [[1, a[1], a[2]],
         [1, b[1], b[2]],
         [1, c[1], c[2]]])
    y = np.linalg.det(
        [[a[0], 1, a[2]],
         [b[0], 1, b[2]],
         [c[0], 1, c[2]]])
    z = np.linalg.det(
        [[a[0], a[1], 1],
         [b[0], b[1], 1],
         [c[0], c[1], 1]])

    magnitude = np.sqrt(x**2 + y**2 + z**2)

    return (x/magnitude, y/magnitude, z/magnitude)


def _poly_area(poly):
    # area of polygon poly
    if len(poly) < 3:
        # not a plane - no area
        return 0
    total = [0.0, 0.0, 0.0]
    N = len(poly)
    for i in range(N):
        vi1 = poly[i]
        vi2 = poly[np.mod((i+1), N)]
        prod = np.cross(vi1, vi2)
        total[0] += prod[0]
        total[1] += prod[1]
        total[2] += prod[2]
    result = np.dot(total, _unit_normal(poly[0], poly[1], poly[2]))
    return np.abs(result/2)



from spharpy.classes.audio import (
    SphericalHarmonicTimeData,
    SphericalHarmonicFrequencyData,
    _atleast_3d_first_dimension)
from spharpy import SphericalHarmonicDefinition
import numpy as np
import numpy.testing as npt


def test_atleast_3d_data():
    """Test extending the shape of 1D and 2D array likes."""

    # Prepend dimension for 2D array
    data = np.ones((4, 4))
    data_3d = _atleast_3d_first_dimension(data)
    assert data_3d.ndim == 3
    assert data_3d.shape == (1, 4, 4)

    # Do not change 3D array
    data_3d = _atleast_3d_first_dimension(np.ones((2, 4, 4)))

    assert data_3d.ndim == 3
    assert data_3d.shape == (2, 4, 4)

    # Prepend two dimensions for 1D array
    data_3d = _atleast_3d_first_dimension(np.array([1, 2, 3, 4]))
    assert data_3d.ndim == 3
    assert data_3d.shape == (1, 1, 4)

    # Convert lists to numpy array
    data = [1, 2, 3, 4]
    data_3d_list = _atleast_3d_first_dimension(data)
    npt.assert_equal(data_3d_list, data_3d)


def test_init_sh_time_data():
    data = np.ones((1, 4, 4))
    times = [1, 2, 3, 4]
    sh_time_data = SphericalHarmonicTimeData(
        data, times,  basis_type='real', normalization='SN3D',
        channel_convention="ACN", condon_shortley=False,
        comment="")
    assert isinstance(sh_time_data, SphericalHarmonicTimeData)
    np.testing.assert_allclose(sh_time_data.time, data)


def test_sh_time_data_from_sh_definition():
    """Test init SphericalHarmonicsSignal."""

    shd = SphericalHarmonicDefinition(
        n_max=0, basis_type="real",
        channel_convention="ACN",
        normalization="N3D",
        condon_shortley=False)

    data = np.ones((1, 4, 4))
    times = [1, 2, 3, 4]

    time_data_def = SphericalHarmonicTimeData.from_definition(
        sh_definition=shd, data=data, times=times)

    assert isinstance(time_data_def, SphericalHarmonicTimeData)

    time_data = SphericalHarmonicTimeData(
            data,
            times, basis_type='real',
            channel_convention='ACN',
            normalization='N3D',
            condon_shortley=False)

    assert time_data == time_data_def


def test_init_sh_frequency_data():
    data = np.ones((1, 4, 4), dtype=complex)
    frequencies = [1, 2, 3, 4]
    sh_freq_data = SphericalHarmonicFrequencyData(
        data, frequencies, basis_type='real', normalization='SN3D',
        channel_convention="ACN", condon_shortley=False,
        comment="")
    assert isinstance(sh_freq_data, SphericalHarmonicFrequencyData)
    np.testing.assert_allclose(sh_freq_data.freq, data)


def test_sh_freq_data_from_sh_definition():
    """Test init SphericalHarmonicsSignal."""

    shd = SphericalHarmonicDefinition(
        n_max=0, basis_type="real",
        channel_convention="ACN",
        normalization="N3D",
        condon_shortley=False)

    data = np.ones((1, 4, 4))
    freqs = [1, 2, 3, 4]

    freq_data_def = SphericalHarmonicFrequencyData.from_definition(
        sh_definition=shd, data=data, frequencies=freqs)

    assert isinstance(freq_data_def, SphericalHarmonicFrequencyData)

    freq_data = SphericalHarmonicFrequencyData(
            data,
            freqs, basis_type='real',
            channel_convention='ACN',
            normalization='N3D',
            condon_shortley=False)

    assert freq_data == freq_data_def



"""Beamforming methods for spherical harmonic signals."""

from .beamforming import (
    dolph_chebyshev_weights,
    rE_max_weights,
    maximum_front_back_ratio_weights,
    normalize_beamforming_weights,
)

__all__ = [
    'dolph_chebyshev_weights',
    'rE_max_weights',
    'maximum_front_back_ratio_weights',
    'normalize_beamforming_weights',
]



"""Tests for sh rotations."""

import numpy as np
import spharpy.transforms as transforms
from spharpy.spherical import spherical_harmonic_basis
import spharpy
from spharpy.transforms import SphericalHarmonicRotation
from spharpy.classes.sh import (
    SphericalHarmonicDefinition, SphericalHarmonics)
import pytest
from pyfar import Coordinates
from spharpy.classes.audio import SphericalHarmonicSignal
import pyfar


def test_rotation_matrix_z_axis_complex():
    rot_angle = np.pi/2
    n_max = 2
    reference = np.diag([1, 1j, 1, -1j, -1, 1j, 1, -1j, -1])

    rot_mat = transforms.rotation_z_axis(n_max, rot_angle)
    np.testing.assert_allclose(rot_mat, reference)

    rot_angle = np.pi
    reference = np.diag([1, -1, 1, -1, 1, -1, 1, -1, 1])

    rot_mat = transforms.rotation_z_axis(n_max, rot_angle)
    np.testing.assert_allclose(rot_mat, reference)

    rot_angle = 3/2*np.pi
    reference = np.diag([1, -1j, 1, 1j, -1, -1j, 1, 1j, -1])

    rot_mat = transforms.rotation_z_axis(n_max, rot_angle)
    np.testing.assert_allclose(rot_mat, reference)


def test_rotation_sh_basis_z_axis_complex():
    rot_angle = np.pi/2
    n_max = 2
    coords_x = Coordinates(1, 0, 0)
    coords_y = Coordinates(0, 1, 0)
    reference = spherical_harmonic_basis(n_max, coords_y).T.conj()

    rot_mat = transforms.rotation_z_axis(n_max, rot_angle)
    sh_vec_x = spherical_harmonic_basis(n_max, coords_x)
    sh_vec_rotated = rot_mat @ sh_vec_x.T.conj()

    sampling = spharpy.samplings.dodecahedron()
    Y = spharpy.spherical.spherical_harmonic_basis(n_max, sampling)

    np.testing.assert_allclose(Y@sh_vec_rotated, Y@reference)


def test_rotation_maxtrix_z_axis_real():
    n_max = 2
    rot_angle = np.pi/4
    reference = np.array([
        [1, 0, 0, 0, 0, 0, 0, 0, 0],
        [0, np.cos(rot_angle), 0, np.sin(rot_angle), 0, 0, 0, 0, 0],
        [0, 0, 1, 0, 0, 0, 0, 0, 0],
        [0, -np.sin(rot_angle), 0, np.cos(rot_angle), 0, 0, 0, 0, 0],
        [0, 0, 0, 0, np.cos(2*rot_angle), 0, 0, 0, np.sin(2*rot_angle)],
        [0, 0, 0, 0, 0, np.cos(rot_angle), 0, np.sin(rot_angle), 0],
        [0, 0, 0, 0, 0, 0, 1, 0, 0],
        [0, 0, 0, 0, 0, -np.sin(rot_angle), 0, np.cos(rot_angle), 0],
        [0, 0, 0, 0, -np.sin(2*rot_angle), 0, 0, 0, np.cos(2*rot_angle)]])
    rot_matrix = transforms.rotation_z_axis_real(n_max, rot_angle)
    np.testing.assert_allclose(rot_matrix, reference)


def test_rotation_sh_basis_z_axis_real():
    rot_angle = np.pi/2
    n_max = 2
    coords_x = Coordinates(1, 0, 0)
    coords_y = Coordinates(0, 1, 0)
    reference = np.squeeze(
        spharpy.spherical.spherical_harmonic_basis_real(n_max, coords_y))

    rot_mat = spharpy.transforms.rotation_z_axis_real(n_max, rot_angle)
    sh_vec_x = spharpy.spherical.spherical_harmonic_basis_real(n_max, coords_x)
    sh_vec_rotated = np.squeeze(rot_mat @ np.squeeze(sh_vec_x))

    np.testing.assert_almost_equal(sh_vec_rotated, np.squeeze(reference))


def test_wigner_d_rot_real():
    euler_angles = []
    n_max = 5
    euler_angles = np.deg2rad(np.array([90., 45., -90.]))
    D = spharpy.transforms.wigner_d_rotation_real(
        n_max, euler_angles[0], euler_angles[1], euler_angles[2])

    reference = np.squeeze(spharpy.spherical.spherical_harmonic_basis_real(
            n_max, Coordinates(0, 1, 1)))

    sh_vec = np.squeeze(spharpy.spherical.spherical_harmonic_basis_real(
        n_max, Coordinates(0, 1, 0)))
    sh_vec_rotated = D @ sh_vec

    np.testing.assert_allclose(sh_vec_rotated, reference, atol=1e-10)

    rot_angle_z = np.pi/4
    D = spharpy.transforms.wigner_d_rotation_real(
        n_max, rot_angle_z, 0, 0)

    rot_mat = spharpy.transforms.rotation_z_axis_real(n_max, rot_angle_z)

    np.testing.assert_allclose(D, rot_mat, atol=1e-7)

    pass


def test_wigner_d_rot():
    euler_angles = []
    n_max = 5
    euler_angles = np.deg2rad(np.array([90., 45., -90.]))
    D = spharpy.transforms.wigner_d_rotation(
        n_max, euler_angles[0], euler_angles[1], euler_angles[2])

    reference = np.squeeze(spharpy.spherical.spherical_harmonic_basis(
            n_max, Coordinates(0, 1, 1)))

    sh_vec = np.squeeze(spharpy.spherical.spherical_harmonic_basis(
        n_max, Coordinates(0, 1, 0)))
    sh_vec_rotated = D @ sh_vec

    np.testing.assert_allclose(sh_vec_rotated, reference)

    rot_angle_z = np.pi/4
    D = spharpy.transforms.wigner_d_rotation(
        n_max, rot_angle_z, 0, 0)

    rot_mat = spharpy.transforms.rotation_z_axis(n_max, rot_angle_z)

    np.testing.assert_allclose(D, rot_mat, atol=1e-7)


def test_SphericalHarmonicRotation():
    """Test SphericalHarmonicRotation creation and check against reference
    for a 90 deg rotation around the z-axis.
    """
    n_max = 2
    definition = SphericalHarmonicDefinition(n_max=n_max)
    rot_angle_z = np.pi/2
    rot_vec = [0, 0, rot_angle_z]
    rot = SphericalHarmonicRotation.from_rotvec(rot_vec)

    D_Rot = rot.as_spherical_harmonic_matrix(definition)

    reference = np.array([
        [1, 0, 0, 0, 0, 0, 0, 0, 0],
        [0, np.cos(rot_angle_z), 0, np.sin(rot_angle_z), 0, 0, 0, 0, 0],
        [0, 0, 1, 0, 0, 0, 0, 0, 0],
        [0, -np.sin(rot_angle_z), 0, np.cos(rot_angle_z), 0, 0, 0, 0, 0],
        [0, 0, 0, 0, np.cos(2*rot_angle_z), 0, 0, 0, np.sin(2*rot_angle_z)],
        [0, 0, 0, 0, 0, np.cos(rot_angle_z), 0, np.sin(rot_angle_z), 0],
        [0, 0, 0, 0, 0, 0, 1, 0, 0],
        [0, 0, 0, 0, 0, -np.sin(rot_angle_z), 0, np.cos(rot_angle_z), 0],
        [0, 0, 0, 0, -np.sin(2*rot_angle_z), 0, 0, 0, np.cos(2*rot_angle_z)]])

    np.testing.assert_allclose(D_Rot, reference, atol=1e-10)

    rot = SphericalHarmonicRotation.from_rotvec([0, 0, np.pi/2])
    np.testing.assert_allclose(
        rot.as_spherical_harmonic_matrix(definition),
        reference, atol=1e-10)

    rot = SphericalHarmonicRotation.from_euler('zyz', [0, 0, np.pi/2])
    np.testing.assert_allclose(
        rot.as_spherical_harmonic_matrix(definition),
        reference, atol=1e-10)

    rot = SphericalHarmonicRotation.from_quat(
        [0, 0, 1/np.sqrt(2), 1/np.sqrt(2)])
    np.testing.assert_allclose(
        rot.as_spherical_harmonic_matrix(definition),
        reference, atol=1e-10)

    rot_mat_z_spat = np.array([
        [0, -1, 0],
        [1, 0, 0],
        [0, 0, 1]])
    rot = SphericalHarmonicRotation.from_matrix(rot_mat_z_spat)
    np.testing.assert_allclose(
        rot.as_spherical_harmonic_matrix(definition),
        reference, atol=1e-7)


def test_SphericalHarmonicRotation_apply():
    """
    Test application of SphericalHarmonicRotation to SphericalHarmonicSignal
    using the apply method, the multiplication operator, and the rotation
    matrix.
    """
    n_max = 2
    spherical_harmonics = SphericalHarmonics(
        n_max=n_max, coordinates=Coordinates(1, 0, 0), inverse_method=None)
    rot_angle_z = np.pi/2
    rot_vec = [0, 0, rot_angle_z]
    rot = SphericalHarmonicRotation.from_rotvec(rot_vec)

    noise = pyfar.signals.noise(512)

    sh_signal = SphericalHarmonicSignal(
        np.atleast_3d(
            spherical_harmonics.basis.T * noise.time).transpose(1, 0, 2),
        noise.sampling_rate,
        basis_type=spherical_harmonics.basis_type,
        normalization=spherical_harmonics.normalization,
        channel_convention=spherical_harmonics.channel_convention,
        condon_shortley=spherical_harmonics.condon_shortley,
    )

    sh_signal_rotated = rot.apply(sh_signal)

    rot_mat = rot.as_spherical_harmonic_matrix(sh_signal)

    sh_data_rotated = rot_mat @ sh_signal._data

    np.testing.assert_allclose(
        sh_signal_rotated._data,
        sh_data_rotated,
        atol=1e-10)

    sh_signal_rot_operator = rot * sh_signal

    np.testing.assert_allclose(
        sh_signal_rotated._data,
        sh_signal_rot_operator._data,
        atol=1e-10)


def test_SphericalHarmonicRotation_mul_rotations():
    """Test multiplication of two SphericalHarmonicRotation objects."""
    rot_angle_z = np.pi/2
    rot_vec = [0, 0, rot_angle_z]
    rot = SphericalHarmonicRotation.from_rotvec(rot_vec)

    result = rot * rot
    result_matrix = result.as_matrix()

    ref = rot.as_matrix() @ rot.as_matrix()

    np.testing.assert_allclose(
        result_matrix,
        ref,
        atol=1e-10)


def test_SphericalHarmonicRotation_mul_pyfar_Rotation():
    """Test multiplying a SphericalHarmonicRotation with a pyfar.Rotation.
    """
    rot_angle_z = np.pi/2
    rot_vec = [0, 0, rot_angle_z]
    rot = SphericalHarmonicRotation.from_rotvec(rot_vec)

    pyfar_rot = pyfar.Rotation.from_rotvec(rot_vec)

    result = rot*pyfar_rot

    assert isinstance(result, SphericalHarmonicRotation)

    ref = rot.as_matrix() @ pyfar_rot.as_matrix()

    np.testing.assert_allclose(
        result.as_matrix(),
        ref,
        atol=1e-10)


def test_SphericalHarmonicRotation_mul_invalid():
    """Test if invalid multiplication operations raise errors."""
    rot = SphericalHarmonicRotation.from_rotvec([0, 0, np.pi/2])

    with pytest.raises(ValueError, match="Multiplication is only supported"):
        rot * 42 # type: ignore


def test_SphericalHarmonicRotation_invalid_definition():
    """Test if invalid spherical harmonic definitions raise errors."""
    rot = SphericalHarmonicRotation.from_rotvec([0, 0, np.pi/2])

    definition = SphericalHarmonicDefinition(
        n_max=2, normalization='N3D', channel_convention='FuMa')
    with pytest.raises(
            NotImplementedError,
            match="Only 'ACN' channel convention is supported",
        ):
        rot.as_spherical_harmonic_matrix(definition)

    definition = SphericalHarmonicDefinition(
        n_max=2, normalization='SN3D', channel_convention='ACN')
    with pytest.raises(
            NotImplementedError,
            match="Only 'N3D' normalization is supported",
        ):
        rot.as_spherical_harmonic_matrix(definition)

